/***************************************************************************
			PRACTICA DE ESTRUCTURA III - Ensamblador

 Autor: Oscar Martnez Valpuesta
 Fecha realizacin: Navidades 1997
 Fecha entrega: 12/01/98

 Objetivo:   Se pretende hallar a partir de un grafo, un subgrafo fuertemente
	   conexo con al menos m nodos y a partir de ah encontrar el camino
	   mnimo entre dos nodos introducidos por el usuario.

 Realizacin:   El grafo inicial tienen n nodos y se genera aleatoriamente
	      inidicando en una matriz las conexiones entre nodos: un 1 en
	      la fila i columna j indica que existe un arco del nodo i al j.
		Para hallar las componentes fuertemente conexas se usa el
	      algoritmo basado en la multiplicacin elemento a elemento de las
	      matrices de alcance y alcanzabilidad.
		Una vez tenemos un subgrafo fuertemente conexo de al menos
	      m nodos, aplicamos el algoritmo de Dijkstra para hallar el
	      camino mnimo entre todos los nodos del subgrafo. La matriz
	      de pesos de los arcos tambin se genera aleatoriamente una
	      vez hallado el subgrafo.

 Nota:   Recordar que en C los vectores parten de la posicin 0. Se va a
       simular las matrices como vectores. Para acceder al elemte (i,j) de
       la matriz se emplear la frmula (i * n) + j, donde n indica el n de
       filas, que nos dar la posicin dentro del vector del mismo elemento.
**************************************************************************/
#include <stdlib.h>	/* rand, srand, malloc */
#include <stdio.h>	/* printf */
#include <time.h>	/* time_t */
#include <conio.h>	/* clrscr */
#include <values.h>	/* MAXINT */

/* macros que nos devuelven la posicin en un vector de un elemento matricial */
#define POS(a,b) (a*nnodos + b)
#define POS2(a,b) (a*mnodos + b)

/******************************	CONSTANTES *********************************/
const INF = MAXINT;		/* 32767 */

/******************************** TIPOS ************************************/
typedef enum {FALSE=0, TRUE=1} boolean;

/****************************** VARIABLES **********************************/
time_t tiempo;	  /* obtener la semilla a partir de la hora */
int *matriz;	  /* matriz de conexin entre nodos */
int *matrizR;	  /* matriz de alcanzabilidad */
int *matrizQ;	  /* matriz de alcance */
int *matrizP; 	  /* matriz de pesos */
int *matrizFC;	  /* matriz de conexin del subgrafo fuertemente conexo */
int *componentes; /* matriz de componentes fuertemente conexas */
int nnodos;       /* n de nodos del grafo */
int mnodos;	  /* n mnimo de nodos del grafo fuertemente conexo */
int maxnodos;	  /* n mximo de nodos del subgrafo fuertemente conexo */
int num_CFC;	  /* n de la CFC con al menos mnodos */
int s;		  /* vrtice inicial */
int t;		  /* vrtice final */
int *pred;	  /* vector que almacena el predecesor de cada nodo */



/****************************** PROTOTIPOS *********************************/
void Pausa(void);
void Introducir(void);
void Generar(void);
void Visualizar(int *m, int n);
void Alcanzabilidad(void);
void CFC(void);
boolean Parametros_correctos(void);
void IniciaD(void);
void Mostrar_camino(int i);
void Dijkstra(void);
extern Inicia(boolean *, int *, int *);
extern Igualar(int *, int *, int *);
extern Calcula_alcance(int *, int *);
extern Calcula_alcanza(int *, int *);
extern Multiplicacion(int *, int *, int *);
extern Buscar(int *, boolean *, int *, int *, int *);
extern Calcula_nodos(int *, int *, int *, int *);
extern Nueva_matriz(int *, int *, int *, int *, int *);
extern Algoritmo1(int *, boolean *, int *, int *, int *);
extern Algoritmo2(int *, boolean *, int *, int *, int *, int *);

/***************************************************************************/
void Pausa(void) {
  printf("\nPausa, pulsa una tecla para continuar...\n");
  getch();
}

/***************************************************************************/
void Introducir(void) {
  clrscr();
  printf("\nIntroduce el nmero de nodos: ");
  scanf("%d", &nnodos);
  printf("Introduce el nmero de nodos del subgrafo fuertemente conexo: ");
  scanf("%d", &mnodos);
}

/***************************************************************************
  Funcin que genera la matriz de conexin entre nodos.
  La matriz indicar con un 0 en la fila i-sima, columna j-sima
  que los nodos i y j no estn unidos por un arco y con 1 que si.
****************************************************************************/
void Generar(void) {
register int i,j;

  /* memoria para la matriz de conexin */
  matriz = (int *) malloc(nnodos * nnodos * sizeof(int));
  if (matriz == NULL) {
    printf("\nError al intentar asignar memoria a la matriz de conexin.");
    exit(0);
  }
  maxnodos = 0;
  /* generacin de las matriz aleatoriamente */
  srand(time(&tiempo)); 	/* semilla, proporcionada por la hora */
  for (i = 0; i < nnodos; i++)
    for (j = 0; j < nnodos; j++) {
      if (i == j)	/* un nodo es autoalcanzable */
	matriz[POS(i,j)]= 1;
      else
	matriz[POS(i,j)]= rand() % 2;  /* enteros entre 0 y 1 */
    }
}
/***************************************************************************
 Visualiza una matriz cuadrada pasada por parmetro. n indica el n de filas
 y columnas.
****************************************************************************/
void Visualizar(int *m, int n) {
register int i,j;

  for (i = 0; i < n; i++) {
    printf("\n");
    for (j = 0; j < n; j++)
      if (n == nnodos)
	printf("%d ", m[POS(i,j)]);
      else
	printf("%d ", m[POS2(i,j)]);
  }
}
/***************************************************************************
 Algoritmo que determina la matriz de alcanzabilidad y la matriz de alcance.

 Matriz de alcanzabilidad: Rij=1 si hay un camino del nodo i al j.
 Algoritmo: se recorre toda la matriz buscando los caminos. Si existe un
 camino del nodo i al j, estar reflejado con un 1 en la fila i, columna j.
 Si existe un camino del nodo j al nodo k, como de i a j tambien hay camino,
 habra que poner un 1 en la fila i, columna k, indicando que del nodo i se
 puede ir al nodo k (a travs de j en este caso).

 Matriz de alcance: Qij=1 si hay un camino de j a i.
 Algoritmo: basta con trasponer la matriz de alcanzabilidad ya que si
 existe un camino de i a j quiere decir que el nodo j es alcanzable desde
 el nodo i, con lo que Qji=1. (Rij = Qji)
****************************************************************************/
void Alcanzabilidad(void) {

  matrizR = (int *) malloc(nnodos * nnodos * sizeof(int));
  matrizQ = (int *) malloc(nnodos * nnodos * sizeof(int));
  if ((matrizR == NULL) || (matrizQ == NULL)) {
    printf("\nError al intentar asignar memoria en las matrices.");
    exit(0);
  }
  /* se igualan las matrices de alcanzabilidad y conexin*/
  Igualar(matrizR, matriz, &nnodos);
  /* calculamos la matriz de alcanzabilidad */
   Calcula_alcanza(matrizR, &nnodos);
  /* se igualan las matrices de alcance y alcanzabilidad */
  Igualar(matrizQ, matrizR, &nnodos);
  /* matriz de alcance, traspuesta de la matriz de alcanzabilidad */
  Calcula_alcance(matrizQ, &nnodos);
}

/***************************************************************************
 Funcin que presenta la matriz de componentes fuertemente conexas en pantalla.
****************************************************************************/
void Presenta(int *m, int n) {
register int i,j;
int cont;	/* n de nodos de la CFC */

  for (i = 0; i < n; i++) {
    printf("\nComponente n %d formada por los nodos: ", i+1);
    cont = 0;
    for (j = 0; j < nnodos; j++)
      if (m[POS(i,j)]) {
	printf("%d ", j+1);
	cont++;
      }
    if (cont > maxnodos) {
      maxnodos = cont;
      num_CFC = i;
    }
  }
}

/***************************************************************************
 Algoritmo que busca las componentes fuertemente conexas. Se basa en la
 multiplicacin elemento a elemento de las matrices de alcanzabilidad y la
 de alcanze. Aquellas filas que sean iguales formarn una componente
 fuertemente conexa. En la matriz resultado estarn almacenadas las
 componentes por filas, indicando con un 1 en la columna correspondiente al
 nodo perteneciente, esto es, que si existe un 1 en la fila i, columna j
 inidica que el nodo j pertenece a la componente i.
****************************************************************************/
void CFC(void) {
int n_comp = 0; 	/* n de componentes */
boolean *usados;	/* indica qu nodos ya pertenecen a una CFC */

  componentes = (int *) calloc(nnodos * nnodos, sizeof(int));
  usados = (boolean *) calloc(nnodos, sizeof(boolean));
  if ((componentes == NULL) || (usados == NULL)) {
    printf("\nError al intentar asignar memoria.");
    exit(0);
  }
  /* inicializacin de usados y de componentes */
  Inicia(usados, componentes, &nnodos);
  /* multiplicacin elemento a elemento de las matrices */
  Multiplicacion(matrizR, matrizQ, &nnodos);
  /* buscamos las componentes */
  Buscar(componentes,usados,matrizQ, &nnodos, &n_comp);

  printf("\nN de CFC: %d", n_comp);
  Presenta(componentes, n_comp);
}

/***************************************************************************
 Funcin que devuelve TRUE si el nodo s y el t estn entre los disponibles.
****************************************************************************/
boolean Parametros_correctos(void) {

  if ((s <= maxnodos) && (s > 0) &&
      (t <= maxnodos) && (t > 0))
    return TRUE;
  printf("\nNodos fuera de rango.");
  return FALSE;
}

/***************************************************************************
 Funcin que construye la nueva matriz de conexin y la matriz de pesos.
****************************************************************************/
void IniciaD(void) {
register int i,j;
int *nodos_pert;	/* nodos pertenecientes a la CFC */

  matrizP = (int *) malloc(mnodos * mnodos * sizeof(int));
  matrizFC = (int *) malloc(mnodos * mnodos * sizeof(int));
  nodos_pert = (int *) malloc(mnodos * sizeof(int));
  if ((matrizP == NULL) || (matrizFC == NULL) || (nodos_pert == NULL)) {
    printf("\nError al intentar asignar memoria.");
    exit(0);
  }
  /* se buscan los nodos pertenecientes a la componente */
  Calcula_nodos(nodos_pert, componentes, &nnodos, &num_CFC);
  /* se construye la nueva matriz de conexin */
  Nueva_matriz(matrizFC, matriz, nodos_pert, &mnodos, &nnodos);
  printf("\nMatrizFC:");
  Visualizar(matrizFC,mnodos);

  /* se genera aleatoriamente la matriz de pesos */
  for (i = 0; i < mnodos; i++)
    for (j = 0; j < mnodos; j++) {
      if ((matrizFC[POS2(i,j)] == 0) || (i == j))  /* si no existe arco */
	matrizP[POS2(i,j)]= 0;	/* el peso es 0 */
      else
	matrizP[POS2(i,j)]= (rand() % 9) + 1;  /* enteros entre 1 y 9 */
    }

  printf("\nMatriz de pesos:");
  Visualizar(matrizP,mnodos);
}

/***************************************************************************
 Funcin que muestra el camino mnimo del nodo s al nodo t.
****************************************************************************/
void Mostrar_camino(int i) {

  if (i != s)
    Mostrar_camino(pred[i]);
  printf("%d ", i+1);
}

/***************************************************************************
 Funcin que realiza el algoritmo de Dijkstra para hallar el camino
 mnimo entre un vrtice y todos los dems.
****************************************************************************/
void Dijkstra(void) {
int *d;		/* vector que guarda la distancia a cada nodo */
boolean *etiquetado;	/* etiquetado temporal */

  do {
    printf("\nIntroduce el vrtice inicial: ");
    scanf("%d", &s);
    printf("Introduce el vrtice final: ");
    scanf("%d", &t);
  } while (!Parametros_correctos());
  s--;
  t--;
  /* memoria para los vectores pred, d y etiquetado */
  pred = (int *) malloc (mnodos * sizeof(int));
  d = (int *) malloc (mnodos * sizeof(int));
  etiquetado = (boolean *) malloc (mnodos * sizeof(boolean));
  if ((pred == NULL) || (d == NULL) || (etiquetado == NULL)) {
    printf("\nError, no se ha podido asignar memoria a los vectores.");
    exit(0);
  }

  /* algoritmo */

  /* se inicia d[i] = INF, etiquetado[i] = FALSE, d[s] = 0, pred[s] = 0 */
  Algoritmo1(d, etiquetado, pred, &s, &mnodos);

  Algoritmo2(d, etiquetado, pred, matrizFC, matrizP, &mnodos);

  printf("\nEl camino mnimo de %d a %d es: ", s+1, t+1);
  Mostrar_camino(t);
  printf("\nDistancia: %d", d[t]);
}

/***************************************************************************/
int main(void) {

  do {
    Introducir();
    Generar();
    printf("\nMatriz del grafo:");
    Visualizar(matriz,nnodos);
    Alcanzabilidad();
    CFC();
    Pausa();
  } while (maxnodos < mnodos);
  mnodos = maxnodos;

  IniciaD();
  Dijkstra();
  Pausa();

  return 0;
}
