/***************************************************************************
   Se genera una matriz con numeros aleatorios que solo pueden ser
 el 0  el 1. La matriz se va a considerar como un vector de tamao nnodos
 por nnodos (ya que tiene que ser cuadrada) donde nnodos es introducido por el
 usuario e inidica el nmero de nodos del grafo. Se le asignar memoria
 dinmicamente. Para acceder al elemento (i,j) de la matriz se usa la
 frmula: (i * n de filas) + j. Recordar que se parte de la posicin 0.
   A partir de ah se calcularn las componentes fuertemente conexas del
 grafo.
***************************************************************************/
#include <stdlib.h>	/* rand, srand, malloc */
#include <stdio.h>	/* printf */
#include <time.h>	/* time_t */
#include <conio.h>	/* clrscr */
#include <values.h>	/* MAXINT */

#define POS(a,b) (a*nnodos + b)
#define POS2(a,b) (a*mnodos + b)

/******************************	CONSTANTES *********************************/
const INF = MAXINT;

/******************************** TIPOS ************************************/
typedef enum {FALSE=0, TRUE=1, USADO=2} boolean;

/****************************** VARIABLES **********************************/
time_t tiempo;	/* obtener la semilla a partir de la hora */
int *matriz;	/* matriz de conexin entre nodos */
int *matrizP; 	/* matriz de pesos */
int *matrizFC;	/* matriz de conexin del grafo fuertemente conexo */
int nnodos; 	/* n de nodos del grafo */
int mnodos; 	/* n de nodos deseados del subgrafo fuertemente conexo */
int maxnodos;   /* n de mximo de nodos del subgrafo fuertemente conexo */
boolean *marca;	  /* vector que inidica qu nodos se han visitado */
int *orden;	  /* vector que indica el orden de visita */
int num;	  /* ndice que indica el orden de numeracin */
int *componentes;	/* matriz de CFC */
int num_CFC;	  /* n de CFC con mnodos */
int s;		/* vrtice inicial */
int t;		/* vrtice final */
int *pred;	/* vector que almacena el predecesor de cada nodo */

/****************************** PROTOTIPOS *********************************/
void Introducir(void);
void Pausa(void);
void Generar(void);
void Visualizar(int *m, int n);
void Iniciar(void);
void dfs(int u, int *m);
void dfs2(int u, int *m);
int Maximo(int *v);
int Maximo2(int *v);
boolean Todos_visitados(void);
void Poner(int *m,int n);
void Presentar(int *m, int n);
void CFC(void);
void IniciaD(void);
void Dijkstra(void);

/***************************************************************************/
void Introducir(void) {
  clrscr();
  printf("\nIntroduce el nmero de nodos: ");
  scanf("%d", &nnodos);
  printf("Introduce el nmero de nodos del subgrafo fuertemente conexo: ");
  scanf("%d", &mnodos);
}

/***************************************************************************/
void Pausa(void) {
  printf("\nPausa, pulsa una tecla para continuar...\n");
  getch();
}

/***************************************************************************
  Funcin que genera la matriz de conexin entre nodos.
  La primera contendr indicar con un 0 en la fila i-sima, columna j-sima
  que los nodos i y j no estn unidos por un arco y con 1 que si.
****************************************************************************/
void Generar(void) {
register int i,j;

  /* memoria para las matrices */
  matriz = (int *) malloc(nnodos * nnodos * sizeof(int));
  if (matriz == NULL) {
    printf("\nError al intentar asignar memoria a la matriz.");
    exit(0);
  }
  maxnodos = 0;
  /* generacin de las matrices aleatoriamente */
  srand((unsigned)time(&tiempo)); 	/* semilla, proporcionada por la hora */
  for (i = 0; i < nnodos; i++)
    for (j = 0; j < nnodos; j++) {
      if (i == j)	/* un nodo es autoalcanzable */
	matriz[POS(i,j)]= 1;
      else
	matriz[POS(i,j)]= rand() % 2;  /* enteros entre 0 y 1 */


    }
}

/***************************************************************************
 Visualiza una matriz determinada.
****************************************************************************/
void Visualizar(int *m, int n) {
register int i,j;

  for (i = 0; i < n; i++) {
    printf("\n");
    for (j = 0; j < n; j++)
      if (n == nnodos)
	printf("%d ", m[POS(i,j)]);
      else
	printf("%d ", m[POS2(i,j)]);
  }
}

/***************************************************************************
 Inicia el vector de marcas como no_visitado (FALSE) y reserva memoria para
 rl vector de orden de visita.
****************************************************************************/
void Iniciar(void) {
register int i;

  marca = (boolean *) malloc(nnodos * sizeof(boolean));
  orden = (int *) malloc(nnodos * sizeof(int));
  for (i = 0; i < nnodos; i++) {
    marca[i] = FALSE;
    orden[i] = nnodos+1;
  }
  num = 1;
}

/***************************************************************************
 Busqueda en profundidad numerando en el orden inverso de visita.
****************************************************************************/
void dfs(int u, int *m) {
int v = 0;

  marca[u] = TRUE;
  while (v <= nnodos) {
    if ((!marca[v]) && (m[POS(u,v)])) {
      dfs(v, m);
      orden[v] = num;
      num++;
    }
    v++;
  }
}

/***************************************************************************
 Busqueda en profundidad.
****************************************************************************/
void dfs2(int u, int *m) {
int v = 0;

  marca[u] = TRUE;
  while (v <= nnodos) {
    if ((!marca[v]) && (m[POS(u,v)]))
      dfs(v, m);
    v++;
  }
}

/***************************************************************************
  Funcin que calcula el vrtice con numeracin ms alta.
****************************************************************************/
int Maximo(int *v) {
register int i;
int maxi = 0;
int vertice = 0;

  for (i = 0; i < nnodos; i++)
    if (v[i] > maxi) {
      maxi = v[i];
      vertice = i;
    }

  return vertice;
}

/***************************************************************************
  Funcin que calcula el vrtice con numeracin ms alta de los no visitados.
****************************************************************************/
int Maximo2(int *v) {
register int i;
int maxi = 0;
int vertice = 0;

  for (i = 0; i < nnodos; i++)
    if ((v[i] > maxi) && (!marca[i])) {
      maxi = v[i];
      vertice = i;
    }

  return vertice;
}

/***************************************************************************
 Funcin que devuelve TRUE si todos los nodos han sido visitados.
****************************************************************************/
boolean Todos_visitados(void) {
register int i;
boolean visitados = TRUE;

  for (i = 0; (i < nnodos) && (visitados); i++)
    if (!marca[i])
      visitados = FALSE;

  return visitados;
}

/***************************************************************************
 Funcin que pone en la matriz m los nodos visitados en la ltima bsqueda
 en profundidad y los marca como usados para no volverlos a poner en otra CFC.
 En la fila i-sima, las columnas que tengan un 1 pertenecen a la CFC i-sima.
****************************************************************************/
void Poner(int *m,int n) {
register int i, cont = 0;

  for (i = 0; i < nnodos; i++)
    if (marca[i] == TRUE) {
      marca[i] = USADO;
      cont++;
      m[POS((n-1),i)] = 1;
    }
  if (cont > maxnodos) {
    maxnodos = cont;
    num_CFC = n;
  }
}

/***************************************************************************
 Funcin que presenta la matriz de componentes en pantalla.
****************************************************************************/
void Presenta(int *m, int n) {
register int i,j;

  for (i = 0; i < n; i++) {
    printf("\nComponente n %d formada por los nodos: ", i+1);
    for (j = 0; j < nnodos; j++)
      if (m[POS(i,j)])
	printf("%d ", j+1);
  }
}


/***************************************************************************
 Algoritmo que busca las componentes fuertemente conexas:
 1. Efectuar un recorrido en profundidad en G y numerar los vrtices en el
    orden de terminacin de las llamadas recursivas.
 2. Construir un nuevo grafo dirigido Gr invirtiendo las direcciones de todos
    los arcos de G.
 3. Realizar un recorrido en profundidad en Gr, partiendo del vrtice con
    numeracin ms alta de acuerdo con la realizada en el paso 1. Si el
    recorrido no llega a todos los vrtices, iniciar el siguiente recorrido
    a partir del vrtice restante con numeracin ms alta.
 4. Cada rbol resultante es una componente fuertemente conexa del grafo G.
****************************************************************************/
void CFC(void) {
register int i,j;
int *Gr;		/* grafo invertido */
int maxi;		/* vrtice con numeracin ms alta */
int n_CFC = 0;		/* n de componentes fuertemente conexas */


  componentes = (int *) malloc(nnodos * nnodos * sizeof(int));
  if (componentes == NULL) {
    printf("\nError al intentar asignar memoria a la matriz de CFC.");
    exit(0);
  }
  for (i = 0; i < nnodos; i++)
    for (j = 0; j < nnodos; j++)
      componentes[POS(i,j)] = 0;

  /* paso 1 */
  Iniciar();
  dfs(0, matriz);
  orden[0] = num;

  /* paso 2 */
  Gr = (int *) malloc(nnodos * nnodos * sizeof(int));
  if (Gr == NULL) {
    printf("\nError al intentar asignar memoria al grafo invertido.");
    exit(0);
  } /* traspones la matriz */
  for (i = 0; i < nnodos; i++)
    for (j = 0; j < nnodos; j++)
      Gr[POS(j,i)] = matriz[POS(i,j)];

  /* paso 3 */
  maxi = Maximo(orden);
  for (i = 0; i < nnodos; i++)
    marca[i] = FALSE;
  dfs2(maxi, Gr);
  n_CFC++;
  Poner(componentes,n_CFC);
  while (!Todos_visitados()) {
    maxi = Maximo2(orden);
    dfs2(maxi,Gr);
    n_CFC++;
    Poner(componentes,n_CFC);
  }
  printf("\nN de CFC: %d", n_CFC);
  Presenta(componentes, n_CFC);
  free(Gr);
  free(marca);
  free(orden);
}

/***************************************************************************/

/***************************************************************************
 Funcin que construye las nuevas matrices de peso y la de conexin.
****************************************************************************/
void IniciaD(void) {
register int i,j;
int *nodos_pert;	/* nodos pertenecientes a la CFC */

  matrizP = (int *) malloc(mnodos * mnodos * sizeof(int));
  matrizFC = (int *) malloc(mnodos * mnodos * sizeof(int));
  nodos_pert = (int *) malloc(mnodos * sizeof(int));
  if ((matrizP == NULL) || (matrizFC == NULL) || (nodos_pert == NULL)) {
    printf("\nError al intentar asignar memoria.");
    exit(0);
  }
  /* se construye la nueva matriz de conexin */
  j = 0;
  for (i = 0; i < nnodos; i++)
    if (componentes[POS((num_CFC-1),i)]) { /* si el nodo i pertenece a la CFC */
      nodos_pert[j] = i;
      j++;
    }
  for (i = 0; i < mnodos; i++)
    for (j = 0; j < mnodos; j++)
      matrizFC[POS2(i,j)] = matriz[POS(nodos_pert[i],nodos_pert[j])];

  /* se genera aleatoriamente la matriz de pesos */
  for (i = 0; i < mnodos; i++)
    for (j = 0; j < mnodos; j++) {
      if ((matrizFC[POS2(i,j)] == 0) || (i == j))  /* si no existe arco */
	matrizP[POS2(i,j)]= 0;	/* el peso es 0 */
      else
	matrizP[POS2(i,j)]= (rand() % 9) + 1;  /* enteros entre 1 y 9 */
    }

  printf("\nMatriz de pesos:");
  Visualizar(matrizP,mnodos);
  free(nodos_pert);
  free(componentes);
  free(matriz);
}

/***************************************************************************
 Funcin que muestra el camino mnimo del nodo s al nodo t.
****************************************************************************/
void Mostrar_camino(int i) {

  if (i != s)
    Mostrar_camino(pred[i]);
  printf("%d ", i+1);
}

/***************************************************************************
 Funcin que realiza el algoritmo de Dijkstra para hallar el camino
 mnimo entre un vrtice y todos los dems.
****************************************************************************/
void Dijkstra(void) {
register iter,i;
int p;		/* nodo auxiliar */
int *d;		/* vector que guarda la distancia a cada nodo */
boolean *etiquetado;	/* etiquetado temporal */
int min;
int candidato;

  printf("\nIntroduce el vrtice inicial: ");
  scanf("%d", &s);
  printf("Introduce el vrtice final: ");
  scanf("%d", &t);
  s--;
  t--;
  /* memoria para los vectores pred, d y etiquetado */
  pred = (int *) malloc (mnodos * sizeof(int));
  d = (int *) malloc (mnodos * sizeof(int));
  etiquetado = (boolean *) malloc (mnodos * sizeof(boolean));
  if ((pred == NULL) || (d == NULL)) {
    printf("\nError, no se ha podido asignar memoria a los vectores.");
    exit(0);
  }

  /* algoritmo */
  for (i = 0; i < mnodos; i++) {
    d[i] = INF;
    etiquetado[i] = FALSE;
  }
  d[s] = 0;
  pred[s] = 0;
  for (iter = 0; iter < mnodos; iter++) {
    min = INF;
    for (i = 0; i < mnodos; i++)
      if ((!etiquetado[i]) && (d[i] < min)) {
	min = d[i];
	candidato = i;
      }
    etiquetado[candidato] = TRUE;
    /* sucesor de candidato */
    for (i = 0; (i < mnodos) && (!matrizFC[POS2(candidato,i)]); i++);
    p = i;
    while (p <= mnodos) {
      if ((!etiquetado[p])&&(d[p] > d[candidato] + matrizP[POS2(candidato,p)])){
	d[p] = d[candidato] + matrizP[POS2(candidato,p)];
	pred[p] = candidato;
      }
      /* sucesor de p */
      p++;
      for (i = p; (i < mnodos) && (!matrizFC[POS2(candidato,i)]); i++);
      p = i;
    }
  }
  printf("\nEl camino mnimo de %d a %d es: ", s+1, t+1);
  Mostrar_camino(t);
  printf("\nDistancia: %d", d[t]);
}
/***************************************************************************/
int main(void) {

  do {
    Introducir();
    Generar();
    printf("\nMatriz del grafo:");
    Visualizar(matriz,nnodos);
    CFC();
    Pausa();
    if (maxnodos < mnodos)
      free(matriz);
  } while (maxnodos < mnodos);
  mnodos = maxnodos;
  IniciaD();
  Dijkstra();
  Pausa();
  return 0;
}