/***************************************************************************
 Se intenta generar una matriz con numeros aleatorios que solo pueden ser
 el 0  el 1. La matriz se va a considerar como un vector de tamao n_fil
 por n_col (n de filas por n de columnas) introducido por el usuario. Se
 le asignar memoria dinmicamente. Para acceder al elemento (i,j) de la
 matriz se usa la frmula: (i * n_fil) + j. Recordar que se parte de la
 posicin 0.
***************************************************************************/
#include <stdlib.h>	/* rand, srand, malloc */
#include <stdio.h>	/* printf */
#include <time.h>	/* time_t */
#include <conio.h>	/* clrscr */
#include <values.h>	/* MAXINT */

#define POS(a,b) (a*nnodos + b)
#define POS2(a,b) (a*mnodos + b)

/******************************	CONSTANTES *********************************/
const INF = MAXINT;

/******************************** TIPOS ************************************/
typedef enum {FALSE=0, TRUE=1} boolean;

/****************************** VARIABLES **********************************/
time_t tiempo;	  /* obtener la semilla a partir de la hora */
int *matriz;	  /* matriz de conexin entre nodos */
int *matrizR;	  /* matriz de alcanzabilidad */
int *matrizQ;	  /* matriz de alcance */
int *matrizP; 	  /* matriz de pesos */
int *matrizFC;	  /* matriz de conexin del subgrafo fuertemente conexo */
int *componentes; /* matriz de componentes fuertemente conexas */
int nnodos;       /* n de nodos del grafo */
int mnodos;	  /* n mnimo de nodos del grafo fuertemente conexo */
int maxnodos;	  /* n mximo de nodos del subgrafo fuertemente conexo */
int num_CFC;	  /* n de la CFC con al menos mnodos */
int s;		  /* vrtice inicial */
int t;		  /* vrtice final */
int *pred;	  /* vector que almacena el predecesor de cada nodo */



/****************************** PROTOTIPOS *********************************/
void Pausa(void);
void Introducir(void);
void Generar(void);
void Visualizar(int *m, int n);
void Alcanzabilidad(void);
void CFC(void);
boolean Parametros_correctos(void);
void IniciaD(void);
void Mostrar_camino(int i);
void Dijkstra(void);

/***************************************************************************/
void Pausa(void) {
  printf("\nPausa, pulsa una tecla para continuar...\n");
  getch();
}

/***************************************************************************/
void Introducir(void) {
  clrscr();
  printf("\nIntroduce el nmero de nodos: ");
  scanf("%d", &nnodos);
  printf("Introduce el nmero de nodos del subgrafo fuertemente conexo: ");
  scanf("%d", &mnodos);
}

/***************************************************************************
  Funcin que genera la matriz de conexin entre nodos.
  La primera contendr indicar con un 0 en la fila i-sima, columna j-sima
  que los nodos i y j no estn unidos por un arco y con 1 que si.
****************************************************************************/
void Generar(void) {
register int i,j;

  /* memoria para la matriz de conexin */
  matriz = (int *) malloc(nnodos * nnodos * sizeof(int));
  if (matriz == NULL) {
    printf("\nError al intentar asignar memoria a la matriz de conexin.");
    exit(0);
  }
  maxnodos = 0;
  /* generacin de las matriz aleatoriamente */
  srand((unsigned) 9); /*time(&tiempo)); 	/* semilla, proporcionada por la hora */
  for (i = 0; i < nnodos; i++)
    for (j = 0; j < nnodos; j++) {
      if (i == j)	/* un nodo es autoalcanzable */
	matriz[POS(i,j)]= 1;
      else
	matriz[POS(i,j)]= rand() % 2;  /* enteros entre 0 y 1 */
    }
}
/***************************************************************************
 Visualiza una matriz determinada.
****************************************************************************/
void Visualizar(int *m, int n) {
register int i,j;

  for (i = 0; i < n; i++) {
    printf("\n");
    for (j = 0; j < n; j++)
      if (n == nnodos)
	printf("%d ", m[POS(i,j)]);
      else
	printf("%d ", m[POS2(i,j)]);
  }
}
/***************************************************************************
 Algoritmo que determina la matriz de alcanzabilidad y la matriz de alcance.
 Matriz de alcanzabilidad: Rij=1 si hay un camino del nodo i al j.
 Algoritmo: se recorre toda la matriz buscando los caminos. Si existe un
 camino del nodo i al j, estar reflejado con un 1 en la fila i, columna j.
 Si existe un camino del nodo j al nodo k, como de i a j tambien hay camino,
 habra que poner un 1 en la fila i, columna k, indicando que del nodo i se
 puede ir al nodo k (a travs de j en este caso).

 Matriz de alcance: Qij=1 si hay un camino de j a i.
 Algoritmo: basta con trasponer la matriz de alcanzabilidad ya que si
 existe un camino de i a j quiere decir que el nodo j es alcanzable desde
 el nodo i, con lo que Qji=1. (Rij = Qji)
****************************************************************************/
void Alcanzabilidad(void) {
register int i,j,k;
boolean cambio = TRUE;


  matrizR = (int *) malloc(nnodos * nnodos * sizeof(int));
  matrizQ = (int *) malloc(nnodos * nnodos * sizeof(int));
  if ((matrizR == NULL) || (matrizQ == NULL)) {
    printf("\nError al intentar asignar memoria en las matrices.");
    exit(0);
  }
  /* se igualan las matrices */
  for (i = 0; i < nnodos; i++)
    for (j = 0; j < nnodos; j++)
      matrizR [POS(i,j)] = matrizQ[POS(i,j)] = matriz[POS(i,j)];

  /* calculamos la matriz de alcanzabilidad */
  while (cambio) {
    cambio = FALSE;
    /* recorro toda la matriz */
    for (i = 0; i < nnodos; i++)
      for (j = 0; j < nnodos; j++)
/* si existe camino de i a j se miran los arcos que salen de j y */
/* se aaden a los de i */
	if ((matrizR[POS(i,j)] == 1) && (i != j))
	  for (k = 0; k < nnodos; k++)
	    if ((matrizR[POS(j,k)] == 1) && (matrizR[POS(i,k)] != 1)) {
	      matrizR[POS(i,k)] = 1; /* se aade el camino de i a k */
	      cambio= TRUE;
	    }
  }

  /* matriz de alcance, traspuesta de la matriz de alcanzabilidad */
  for (i = 0; i < nnodos; i++)
    for (j = 0; j < nnodos; j++)
	matrizQ[POS(j,i)] = matrizR[POS(i,j)];
}

/***************************************************************************
 Funcin que presenta la matriz de componentes en pantalla.
****************************************************************************/
void Presenta(int *m, int n) {
register int i,j;
int cont;	/* n de nodos de la CFC */

  for (i = 0; i < n; i++) {
    printf("\nComponente n %d formada por los nodos: ", i+1);
    cont = 0;
    for (j = 0; j < nnodos; j++)
      if (m[POS(i,j)]) {
	printf("%d ", j+1);
	cont++;
      }
    if (cont > maxnodos) {
      maxnodos = cont;
      num_CFC = i;
    }
  }
}

/***************************************************************************
 Algoritmo que busca las componentes fuertemente conexas. Se basa en la
 multiplicacin elemento a elemento de las matrices de alcanzabilidad y la
 de alcanze. Aquellas filas que sean iguales formarn una componente
 fuertemente conexa. En la matriz resultado estarn almacenadas las
 componentes por filas, indicando con un 1 en la columna correspondiente al
 nodo perteneciente.
****************************************************************************/
void CFC(void) {
int *aux; 	/* matriz auxiliar */
int n_comp = 0; /* n de componentes */
boolean *usados;	/* indica que nodos ya pertenecen a una CFC */
boolean aumento;
register int i,j;

  aux = (int *) malloc(nnodos * nnodos * sizeof(int));
  usados = (boolean *) malloc(nnodos * sizeof(boolean));
  componentes = (int *) malloc(nnodos * nnodos * sizeof(int));
  if ((aux == NULL) || (componentes == NULL) || (usados == NULL)) {
    printf("\nError al intentar asignar memoria.");
    exit(0);
  }
  /* multiplicacin elemento a elemento de las matrices */
  for (i = 0; i < nnodos; i++) {
    usados[i] = FALSE;
    for (j = 0; j < nnodos; j++) {
      aux[POS(i,j)] = matrizR[POS(i,j)] * matrizQ[POS(i,j)];
      componentes[POS(i,j)] = 0;
    }
  }

  /* buscamos las componentes */
  for (i = 0; i < nnodos; i++) {
    aumento = FALSE;
    for (j = 0; j < nnodos; j++)
      if ((aux[POS(i,j)]) && (!usados[j])) {
	componentes[POS(n_comp,j)] = 1;
	usados[j] = TRUE;
	aumento = TRUE;
      }
    if (aumento)
      n_comp++;
  }
  printf("\nN de CFC: %d", n_comp);
  Presenta(componentes, n_comp);
  free(matrizR);
  free(matrizQ);
  free(aux);
  free(usados);
}

/***************************************************************************
 Funcin que devuelve TRUE si el nodo s y el t estn entre los disponibles.
****************************************************************************/
boolean Parametros_correctos(void) {

  if ((s <= maxnodos) && (s > 0) &&
      (t <= maxnodos) && (t > 0))
    return TRUE;
  printf("\nNodos fuera de rango.");
  return FALSE;
}

/***************************************************************************
 Funcin que construye las nueva matriz de conexin y la matriz de pesos.
****************************************************************************/
void IniciaD(void) {
register int i,j;
int *nodos_pert;	/* nodos pertenecientes a la CFC */

  matrizP = (int *) malloc(mnodos * mnodos * sizeof(int));
  matrizFC = (int *) malloc(mnodos * mnodos * sizeof(int));
  nodos_pert = (int *) malloc(mnodos * sizeof(int));
  if ((matrizP == NULL) || (matrizFC == NULL) || (nodos_pert == NULL)) {
    printf("\nError al intentar asignar memoria.");
    exit(0);
  }
  /* se construye la nueva matriz de conexin */
  j = 0;
  for (i = 0; i < nnodos; i++)
    if (componentes[POS(num_CFC,i)]) { /* si el nodo i pertenece a la CFC */
      nodos_pert[j] = i;
      j++;
    }
  for (i = 0; i < mnodos; i++)
    for (j = 0; j < mnodos; j++)
      matrizFC[POS2(i,j)] = matriz[POS(nodos_pert[i],nodos_pert[j])];
  printf("\nMatrizFC:");
  Visualizar(matrizFC,mnodos);

  /* se genera aleatoriamente la matriz de pesos */
  for (i = 0; i < mnodos; i++)
    for (j = 0; j < mnodos; j++) {
      if ((matrizFC[POS2(i,j)] == 0) || (i == j))  /* si no existe arco */
	matrizP[POS2(i,j)]= 0;	/* el peso es 0 */
      else
	matrizP[POS2(i,j)]= (rand() % 9) + 1;  /* enteros entre 1 y 9 */
    }

  printf("\nMatriz de pesos:");
  Visualizar(matrizP,mnodos);
  free(nodos_pert);
  free(componentes);
  free(matriz);
}

/***************************************************************************
 Funcin que muestra el camino mnimo del nodo s al nodo t.
****************************************************************************/
void Mostrar_camino(int i) {

  if (i != s)
    Mostrar_camino(pred[i]);
  printf("%d ", i+1);
}

/***************************************************************************
 Funcin que realiza el algoritmo de Dijkstra para hallar el camino
 mnimo entre un vrtice y todos los dems.
****************************************************************************/
void Dijkstra(void) {
register iter,i;
int p;		/* nodo auxiliar */
int *d;		/* vector que guarda la distancia a cada nodo */
boolean *etiquetado;	/* etiquetado temporal */
int min;
int candidato;

  do {
    printf("\nIntroduce el vrtice inicial: ");
    scanf("%d", &s);
    printf("Introduce el vrtice final: ");
    scanf("%d", &t);
  } while (!Parametros_correctos());
  s--;
  t--;
  /* memoria para los vectores pred, d y etiquetado */
  pred = (int *) malloc (mnodos * sizeof(int));
  d = (int *) malloc (mnodos * sizeof(int));
  etiquetado = (boolean *) malloc (mnodos * sizeof(boolean));
  if ((pred == NULL) || (d == NULL)) {
    printf("\nError, no se ha podido asignar memoria a los vectores.");
    exit(0);
  }

  /* algoritmo */
  for (i = 0; i < mnodos; i++) {
    d[i] = INF;
    etiquetado[i] = FALSE;
  }
  d[s] = 0;
  pred[s] = 0;
/*   candidato = s;
  while (candidato != t) { */
  for (iter = 0; iter < mnodos; iter++) {
    min = INF;
    for (i = 0; i < mnodos; i++)
      if ((!etiquetado[i]) && (d[i] < min)) {
	min = d[i];
	candidato = i;
      }
    etiquetado[candidato] = TRUE;
    /* sucesor de candidato */
    for (i = 0; (i < mnodos) && (!matrizFC[POS2(candidato,i)]); i++);
    p = i;
    while (p <= mnodos) {
      if ((!etiquetado[p])&&(d[p] > d[candidato] + matrizP[POS2(candidato,p)])){
	d[p] = d[candidato] + matrizP[POS2(candidato,p)];
	pred[p] = candidato;
      }
      /* sucesor de p */
      p++;
      for (i = p; (i < mnodos) && (!matrizFC[POS2(candidato,i)]); i++);
      p = i;
    }
  }
  printf("\nEl camino mnimo de %d a %d es: ", s+1, t+1);
  Mostrar_camino(t);
  printf("\nDistancia: %d", d[t]);
}

/***************************************************************************/
int main(void) {

  do {
    Introducir();
    Generar();
    printf("\nMatriz del grafo:");
    Visualizar(matriz,nnodos);
    Alcanzabilidad();
    CFC();
    Pausa();
    if (maxnodos < mnodos)
      free(matriz);
  } while (maxnodos < mnodos);
  mnodos = maxnodos;

  IniciaD();
  Dijkstra();
  Pausa();
  free(matrizP);
  free(matrizFC);

  return 0;
}
